/* ipwatch 
 *
 * ipwatch -- Look at the traffic going by on the ethernet.
 *
 * (1) Need to be root to use this.
 * (2) Uses etherd, nit, mgt, src
 * (3) Defaults to le0 (sparc default)
 * 
 * Written by:
 *
 * 1.0 thd@ornl.gov
 *
 * 1.3 ivan@ccs.neu.edu
 *     Ivan R Judson
 *     Northeastern University
 *     10/13/93
 * --Summary of Modifications--
 * Made it take a -i <interface> switch
 * Added support for getting list of services from NIS.
 *
 * God I hate curses, but if we have to use it...
 *
 * Possibilities:
 *  Tcl/Tk version?
 *  Summary? 
 *  Command line Parsing?
 *  Man Page?
 */

#include "ipwatch.h"

main(argc,argv)
     char	**argv;
{
  void		quit();
  int fd, packet_ip();
  int last, readbit, readfds;
  
  delay=5;
  
  if((argc > 3) || (argc == 2)) 
    printf("usage: ipwatch [-i interface]\n");
  else
    device = (argc == 3)?argv[2]:"le0";
  
  initdevice();
  initscr();
  signal(SIGINT, quit);
  signal(SIGTERM, quit);
  signal(SIGQUIT, quit);
  signal(SIGHUP, quit);
  clear();
  refresh();
  last = time(0);
  readbit = 1 << if_fd;
  while(1) {
    readfds = readbit;
    if (select(32, &readfds, 0, 0, 0) < 0) {
      if (errno == EINTR)
	continue;
      perror("ipwatch: select");
      quit();
    }
    if (readfds & readbit) readdata();
    if ((time(0) - last ) > delay) {
      update();
      last = time(0);
    }
  }
}

update()
{
  int percent, row, i,j, ndots, now;
  row = 0;
  now = time(0);
  move(0, 12);
  printw(" IpWatch-1.3\n\t  Interface %s\n\t ippkts %5d\n", 
	 device, Pkts);
  if (Pkts == 0) {
    refresh(); return;
  }
  row = 1;
  move(row,0);  clrtoeol();
  printw("Load:\n\t %4.0f KBs\n\t %d pps\n", KBs/(1000.*delay), AllPkts/delay);
  move(row, 25); 
  percent = 100.*KBs/(1000.*delay*1200);
  printw("%3d%%",percent);      move(row, 30);
  ndots = percent/2;            while (ndots--) addch('*');
  move(++row, 0);               clrtoeol();
  printw(" arps");              move(row, 16);
  printw("%5d",arps);           move(row, 25);
  printw("%3d",arps/delay);     move(row, 30);
  ndots = arps/delay;
  if (ndots > 50) ndots = 50;
  while (ndots--) addch('*');
  move(++row, 0);               clrtoeol();
  printw(" IPmcast");           move(row, 16);
  printw("%5d",ipmcasts);       move(row, 25);
  printw("%3d",ipmcasts/delay); move(row, 30);
  ndots = ipmcasts/delay;
  if (ndots > 50) ndots=50;
  while (ndots--) addch('*');
  row = 3;
  for (i=0;i<Ntcp;i++) {
    move(++row, 0);
    clrtoeol();                 if (i) move(row,1);
    printw("%s",tcp[i].name);   move(row, 16);
    printw("%5d",tcp[i].count); move(row, 25);
    percent = 100.*tcp[i].count/Pkts;
    printw("%3d%%",percent);    move(row, 30);
    ndots = 50.*tcp[i].count/Pkts;
    while (ndots--) addch('*');
  }
  row++;
  for (i=0;i<Nudp;i++) {
    move(++row, 0);             clrtoeol();
    if (i) move(row,1);
    printw("%s",udp[i].name);   move(row, 16);
    printw("%5d",udp[i].count); move(row, 25);
    percent = 100.*udp[i].count/Pkts;
    printw("%3d%%",percent);    move(row, 30);
    ndots = 50.*udp[i].count/Pkts;
    while (ndots--) addch('*');
  }

  refresh();
  arps = AllPkts = Pkts= ipmcasts=0;
  KBs = 0;
  for (i=0;i<Ntcp;i++) tcp[i].count=0;
  for (i=0;i<Nudp;i++) udp[i].count=0;
}


void quit()
{
  move(23,0);
  refresh();
  endwin();
  sleep(1);
  exit(0);
}

readdata()
{
  register struct nit_bufhdr *hdrp;
  register struct etherstat *p;
  register struct etheraddrs *q;
  caddr_t bp, bufstop;
  char *packet;
  int cc;
  
  if ((cc = read(if_fd, buf, sizeof buf)) < 0) {
    perror("ipwatch: read");
    exit(1);
  }
  
  for (bp = buf, bufstop = buf + cc; bp < bufstop;
       bp += hdrp->nhb_totlen) {

    hdrp = (struct nit_bufhdr *)bp;
    packet = (char *)(hdrp + 1);

    bcopy(packet, (char *)&eheader, sizeof (eheader));
    bcopy(packet + sizeof (eheader),  (char *)&netheader, sizeof (netheader));

    wirelen = hdrp->nhb_msglen;

    edst = (char *)&eheader.ether_dhost;
    esrc = (char *)&eheader.ether_shost;
    
    if (wirelen < MINPACKETLEN) {
      continue;
    }

    AllPkts++;
    KBs = KBs + wirelen;
    
    if (TYPE == ETHERTYPE_ARP)
      ; /* packet_arp(); */
    else
      if (TYPE == ETHERTYPE_IP)
	packet_ip(IP);
    else
      ;      /* packet_pup(); */
  }  
}

initdevice()
{
  struct strioctl si;
  struct ifreq    ifr;
  struct timeval  timeout;
  u_long          flags;
  
  
  if ((if_fd = open("/dev/nit", O_RDONLY)) < 0) {
    perror("nit open");
    exit (1);
  }
  
  /*
   * Arrange to get discrete messages from the stream.
   */
  if (ioctl(if_fd, I_SRDOPT, (char *)RMSGD) < 0) {
    perror("ioctl (I_SRDOPT)");
    exit (1);
  }
  
  si.ic_timout = INFTIM;
  
  /*
   * Push and configure the buffering module.
   */
  if (ioctl(if_fd, I_PUSH, "nbuf") < 0) {
    perror("ioctl (I_PUSH \"nbuf\")");
    exit (1);
  }
  
  timeout.tv_sec = 1;
  timeout.tv_usec = 0;
  si.ic_cmd = NIOCSTIME;
  si.ic_len = sizeof timeout;
  si.ic_dp = (char *)&timeout;
  if (ioctl(if_fd, I_STR, (char *)&si) < 0) {
    perror("ioctl (I_STR: NIOCSTIME)");
    exit (1);
  }
  
  si.ic_cmd = NIOCSCHUNK;
  si.ic_len = sizeof chunksize;
  si.ic_dp = (char *)&chunksize;
  if (ioctl(if_fd, I_STR, (char *)&si) < 0) {
    perror("ioctl (I_STR: NIOCSCHUNK)");
    exit (1);
  }
  
  /*
   * Configure the nit device, binding it to the proper
   * underlying interface and setting the interface into
   * promiscuous mode if necessary.
   */
  (void) strncpy(ifr.ifr_name, device, sizeof ifr.ifr_name);
  ifr.ifr_name[sizeof ifr.ifr_name - 1] = '\0';
  si.ic_cmd = NIOCBIND;
  si.ic_len = sizeof ifr;
  si.ic_dp = (char *)&ifr;
  if (ioctl(if_fd, I_STR, (char *)&si) < 0) {
    perror("ioctl (I_STR: NIOCBIND)");
    exit(1);
  }
  
  flags = NI_PROMISC;
  si.ic_cmd = NIOCSFLAGS;
  si.ic_len = sizeof flags;
  si.ic_dp = (char *)&flags;
  if (ioctl(if_fd, I_STR, (char *)&si) < 0) {
    perror("ioctl (I_STR: NIOCSFLAGS)");
    exit (1);
  }
  
  /*
   * Flush the read queue, to get rid of anything that
   * accumulated before the device reached its final configuration.
   */
  if (ioctl(if_fd, I_FLUSH, (char *)FLUSHR) < 0) {
    perror("ioctl (I_FLUSH)");
    exit(1);
  }
}

packet_ip(ip)
     register struct ip *ip;
{
  struct tcphdr *tp;
  struct udphdr *up;
  int i,fragmented;
  
  Pkts++;
  tp = (struct tcphdr *)(ip + 1);
  up = (struct udphdr *)(ip + 1);
  fragmented = ip->ip_off & 0x1fff;  /*non zero offset */
  if (edst[0] == 1 && edst[1] == 0 && edst[2] == 0x5e ) ipmcasts++;
  
  switch (ip->ip_p) {
  case IPPROTO_TCP:
    tcp[0].count++;
    for(i=1;i< Ntcp-1; i++) {
      if (tp->th_sport == tcp[i].port ||
	  tp->th_dport == tcp[i].port ){
	tcp[i].count++;
	break;
      }
    }
    if (i==(Ntcp-1)) tcp[i].count++;
    break;
  case IPPROTO_UDP:
    udp[0].count++;
    if (fragmented) up->uh_sport = 2049; /*nfs*/
    for(i=1;i< Nudp-1; i++) {
      if (up->uh_sport == udp[i].port ||
	  up->uh_dport == udp[i].port ){
	udp[i].count++;
	break;
      }
    }
    if (i==(Nudp-1)) udp[i].count++;
    break;
  default:
    break;
  }
  return(1);
}
